;; Generated package description from latex-table-wizard.el  -*- no-byte-compile: t; lexical-binding:t -*-
(define-package "latex-table-wizard" "1.5.5" "Magic editing of LaTeX tables" '((emacs "27.1") (auctex "12.1") (transient "0.3.7")) :commit "fcdc34960ec08640adf8019185927f079a2bc89d" :authors '(("Enrico Flor" . "enrico@eflor.net")) :maintainer '("Enrico Flor" . "enrico@eflor.net") :keywords '("convenience" "tex") :url "https://github.com/enricoflor/latex-table-wizard")
